<?php

namespace Drp\NativeEventApp\Modules;

use Drp\NativeEventApp\Exceptions\InvalidModule;
use Drp\NativeEventApp\Exceptions\ModuleNotFound;
use Drp\NativeEventApp\Facades\ModuleRegistry;
use Illuminate\Container\Container as IlluminateContainer;
use Illuminate\Support\Arr;
use Illuminate\Support\Collection;

class ModulesUnSerializer
{
    /** @var IlluminateContainer */
    private $container;

    public function __construct(IlluminateContainer $container)
    {
        $this->container = $container;
    }

    public function getModulesFromString(string $modules): Modules
    {
        return $this->getModulesFromArray(json_decode($modules, true) ?? []);
    }

    public function getModulesFromArray(array $modules): Modules
    {
        return collect($modules)
            ->map(function (array $module) {
                return $this->unserialize($module);
            })
            ->pipe(function (Collection $modules) {
                return Modules::make($modules->toArray());
            });
    }

    private function unserialize(array $moduleData): Module
    {
        try {
            $module = Arr::get(ModuleRegistry::get(), $moduleData['type']);

            throw_if($module === null, ModuleNotFound::class);

            return $this->makeFromContent($moduleData, $module);
        } catch (\Throwable $exception) {
            report(InvalidModule::failedUnserialization($moduleData, $exception));

            return $this->createMissingModule($moduleData);
        }
    }

    private function makeFromContent(array $moduleData, string $moduleName): Module
    {
        $moduleData = $this->canHaveChildren($moduleName) ? $this->transformChildren($moduleData) : $moduleData;

        return $this->container
            ->make($moduleName, $moduleData)
            ->setAttributes($moduleData);
    }

    /**
     * @param object|string $module
     */
    public function canHaveChildren($module): bool
    {
        return is_a($module, Containerable::class, true);
    }

    private function setChildren(Module $module, array $moduleData): Module
    {
        $children = $this->unserializeChildren(Arr::get($moduleData, 'children', []));

        if ($children !== null) {
            $module->children($children);
        }

        return $module;
    }

    private function transformChildren(array $moduleData): array
    {
        $moduleData['children'] = $this->unserializeChildren($moduleData['children']);

        return $moduleData;
    }

    private function unserializeChildren(array $children): ?Modules
    {
        return Modules::fromArray($children);
    }

    private function createMissingModule(array $moduleData): Module
    {
        $module = MissingModule::make()->setAttributes($moduleData);

        return $this->setChildren($module, $moduleData);
    }
}
