<?php

namespace Drp\NativeEventApp;

use DateInterval;
use Drp\NativeEventApp\Auth\AuthInfoFinder;
use Drp\NativeEventApp\Auth\AuthTokenGrant;
use Drp\NativeEventApp\Auth\CodeGenerator;
use Drp\NativeEventApp\Broadcasters\SocketBroadcaster;
use Drp\NativeEventApp\Console\Commands\SendScheduledNotifications;
use Drp\NativeEventApp\Contracts\MagicLinkGenerator;
use Drp\NativeEventApp\Contracts\MagicLinkSigner;
use Drp\NativeEventApp\Contracts\QrCode;
use Drp\NativeEventApp\Contracts\UserAgent;
use Drp\NativeEventApp\Enums\AuthMethod;
use Drp\NativeEventApp\Enums\ModuleType;
use Drp\NativeEventApp\Facades\ModuleRegistry;
use Drp\NativeEventApp\Forms\FormRegistry;
use Drp\NativeEventApp\GraphQL\ContextFactory;
use Drp\NativeEventApp\Http\Middleware\Authenticate;
use Drp\NativeEventApp\Http\ViewComposers\MagicLink;
use Drp\NativeEventApp\Models\Event;
use Drp\NativeEventApp\Models\FileUpload;
use Drp\NativeEventApp\Models\Page;
use Drp\NativeEventApp\Modules\ModulesUnSerializer;
use Drp\NativeEventApp\Notifications\Events\NotificationSendFailed;
use Drp\NativeEventApp\Notifications\Events\NotificationSendSuccess;
use Drp\NativeEventApp\Notifications\Events\SendNotification;
use Drp\NativeEventApp\Notifications\Events\SendPushNotification;
use Drp\NativeEventApp\Notifications\Listeners\NotificationSender;
use Drp\NativeEventApp\Notifications\Listeners\SendPushNotificationToDevice;
use Drp\NativeEventApp\Notifications\Listeners\UpdateNotificationOnFailed;
use Drp\NativeEventApp\Notifications\Listeners\UpdateNotificationOnSuccess;
use Drp\NativeEventApp\Quiz\ApiQuizOperator;
use Drp\NativeEventApp\Quiz\Client\QuizClient;
use Drp\NativeEventApp\Quiz\QuizOperatorManager;
use Drp\NativeEventApp\Templates\DataProviders\DataProviderRegistry;
use Faker\Factory as FakerFactory;
use Faker\Generator;
use GuzzleHttp\Client;
use Illuminate\Console\Scheduling\Schedule;
use Illuminate\Container\Container;
use Illuminate\Database\Eloquent\Factory;
use Illuminate\Database\Eloquent\Relations\Relation;
use Illuminate\Foundation\Http\FormRequest;
use Illuminate\Routing\Router;
use Illuminate\Support\Facades\Event as EventFacade;
use Illuminate\Support\Facades\File;
use Illuminate\Support\Facades\View;
use Illuminate\Support\ServiceProvider as IlluminateServiceProvider;
use Jenssegers\Agent\Agent;
use Laravel\Passport\Bridge\AuthCodeRepository;
use Laravel\Passport\Bridge\RefreshTokenRepository;
use Laravel\Passport\Passport;
use League\OAuth2\Server\AuthorizationServer;
use Nuwave\Lighthouse\Schema\TypeRegistry;
use Nuwave\Lighthouse\Schema\Types\LaravelEnumType;
use Nuwave\Lighthouse\Subscriptions\BroadcastManager;
use Nuwave\Lighthouse\Support\Contracts\CreatesContext;
use SimpleSoftwareIO\QrCode\BaconQrCodeGenerator;

class ServiceProvider extends IlluminateServiceProvider
{
    public function register(): void
    {
        $configPath = __DIR__ . '/../config/native-event-app.php';
        $this->mergeConfigFrom($configPath, 'native-event-app');
        $this->loadViewsFrom(__DIR__ . '/Http/views', 'NativeEventApp');
        $this->loadMigrationsFrom(__DIR__ . '/../database/migrations');

        View::composer('NativeEventApp::magic', MagicLink::class);

        $this->app->singleton(AuthInfoFinder::class);

        $this->registerEloquentFactory();
        $this->registerQuizOperator();

        Relation::morphMap([
            'events' => Event::class,
            'pages' => Page::class,
        ]);

        $this->app->singleton(ModulesUnSerializer::class, static function (Container $app) {
            return new ModulesUnSerializer($app);
        });

        $this->app->singleton(DataProviderRegistry::class, function () {
            return new DataProviderRegistry();
        });

        $this->app->afterResolving(FormRequest::class, function (FormRequest $request, $container): void {
            if ($container->bound('eventpal-form.data') === false) {
                return;
            }

            if (is_array($container->make('eventpal-form.data')) === false) {
                return;
            }

            $request->merge($container->make('eventpal-form.data'));
        });

        $this->app->singleton(FormRegistry::class);
    }

    public function boot(TypeRegistry $typeRegistry, Router $router): void
    {
        $router->middlewareGroup('eventpal.api.auth', [Authenticate::class]);

        $this->registerModelBindings($router);

        $this->app->singleton(CreatesContext::class, ContextFactory::class);
        $this->loadTranslationsFrom(__DIR__ . '/../lang', 'NativeEventApp');

        app(BroadcastManager::class)->extend('socket', function ($app) {
            return $app->make(SocketBroadcaster::class);
        });

        $this->registerOauthServer();
        $this->registerEnums($typeRegistry);
        $this->registerMagicLink();
        $this->registerEvents();

        if (File::exists($this->getRoutePath())) {
            require_once $this->getRoutePath();
        }

        require __DIR__ . '/../routes/file-uploads.php';

        if ($this->app->runningInConsole() === false) {
            return;
        }

        $this->registerCommands();
        $this->registerPublishes();

        $this->app->afterResolving(Schedule::class, static function (Schedule $schedule): void {
            $schedule->command(SendScheduledNotifications::class)->everyMinute();
        });
    }

    private function registerModelBindings(Router $router): void
    {
        $router->bind('eventpalFileUpload', static function (string $uuid) {
            return FileUpload::where('uuid', $uuid)->firstOrFail();
        });
    }

    private function registerEnums(TypeRegistry $typeRegistry): void
    {
        $typeRegistry
            ->register(
                new LaravelEnumType(AuthMethod::class)
            )
            ->register(
                new ModuleType(ModuleRegistry::modulesForEnums())
            );
    }

    private function registerEloquentFactory(): void
    {
        $this->app->singleton(Generator::class, function ($app) {
            return FakerFactory::create($app['config']->get('app.faker_locale', 'en_GB'));
        });

        $this->app->singleton('SeededFaker', function ($app) {
            $faker = $app->make(Generator::class);
            $faker->seed('NEA');

            return $faker;
        });

        $this->app->singleton(Factory::class, function ($app) {
            return Factory::construct(
                $app['SeededFaker'],
                database_path('factories')
            );
        });
    }

    private function registerOauthServer(): void
    {
        $tokenExpiry = new DateInterval(config('native-event-app.access_token.ttl'));
        $refreshTokenExpiry = new DateInterval('P1M');

        $this->app->bind(AuthTokenGrant::class, function () use ($refreshTokenExpiry, $tokenExpiry) {
            return new AuthTokenGrant(
                $this->app->make(AuthCodeRepository::class),
                $this->app->make(RefreshTokenRepository::class),
                $tokenExpiry,
                $refreshTokenExpiry
            );
        });

        $this->app->afterResolving(AuthorizationServer::class, function ($server) use ($tokenExpiry): void {
            $server->enableGrantType($this->makeAuthTokenGrant(), $tokenExpiry);
        });

        Passport::tokensCan([
            'all' => 'All',
        ]);

        Passport::setDefaultScope([
            'all',
        ]);
    }

    private function makeAuthTokenGrant(): AuthTokenGrant
    {
        return tap($this->app->make(AuthTokenGrant::class), function ($grant): void {
            $grant->setRefreshTokenTTL(Passport::refreshTokensExpireIn());
        });
    }

    private function registerMagicLink(): void
    {
        $this->app->bind(QrCode::class, function ($app) {
            return new Auth\QrCode($app->make(BaconQrCodeGenerator::class));
        });
        $this->app->bind(UserAgent::class, function ($app) {
            return new Auth\UserAgent($app->make(Agent::class));
        });

        $this->app->bind(MagicLinkGenerator::class, function ($app) {
            return new Auth\MagicLinkGenerator(
                $app->make(MagicLinkSigner::class),
                $app->make(QrCode::class),
                url(config('lighthouse.route.uri')),
                new DateInterval(config('native-event-app.qr_code.ttl'))
            );
        });

        $this->app->bind(MagicLinkSigner::class, function () {
            return new Auth\MagicLinkSigner(config('app.key'));
        });

        $this->app->bind(CodeGenerator::class, function () {
            return new CodeGenerator('ABCDEFGHJKLMNPQRSTUVWXYZ234567');
        });
    }

    private function registerQuizOperator(): void
    {
        $this->app->bind(QuizOperatorManager::class, function ($app) {
            return new QuizOperatorManager($app);
        });

        $this->app->bind(ApiQuizOperator::class, function () {
            return new ApiQuizOperator(
                new QuizClient(
                    new Client(['base_uri' => config('native-event-app.socket_server.host')])
                )
            );
        });
    }

    private function registerEvents(): void
    {
        EventFacade::listen(SendNotification::class, NotificationSender::class);
        EventFacade::listen(SendPushNotification::class, SendPushNotificationToDevice::class);
        EventFacade::listen(NotificationSendSuccess::class, UpdateNotificationOnSuccess::class);
        EventFacade::listen(NotificationSendFailed::class, UpdateNotificationOnFailed::class);
    }

    private function registerCommands(): void
    {
        $this->commands([
            SendScheduledNotifications::class,
        ]);
    }

    private function registerPublishes(): void
    {
        $this->publishes([
            __DIR__ . '/../config/native-event-app.php' => $this->getConfigPath(),
        ], 'config');

        $this->publishes([
            __DIR__ . '/../database/default-migrations' => database_path('migrations'),
        ], 'migrations');

        $this->publishes([
            __DIR__ . '/../routes/form-submissions.php' => $this->getRoutePath(),
        ], 'routes');
    }

    private function getConfigPath(): string
    {
        return config_path('native-event-app.php');
    }

    private function getRoutePath(): string
    {
        return base_path('routes/form-submissions.php');
    }
}
